'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

from guts.messages import message, infotip, articled
from guts.game_data import game
from guts.utils import coinflip, random_pos, roll_dice, random_choice, eight_directions
from guts.ai import Confusion
from guts.task_resolution import resolve_task, trait_gains


def closest_agent(x, y, max_range = None):
    current_distance = 1
    if max_range == None:
        max_range = max(game.board.width, game.board.height) 
    while current_distance <= max_range:
        agents = [tile.occupant for tile in game.board.ring(x, y, current_distance) if tile.visible and tile.occupant is not None]
        if agents:
            return random_choice(agents)
        current_distance += 1
    return None

def mark_agent(action, agent, target):
    pass

def blow_up(action, agent, target):
    message("The {0} explodes, razing the ground!".format(action.piece.name), 'fiery')
    if target.occupant is not None:
        damage, advantages = _attack(action, agent, target.occupant)
        damage += advantages
        if damage > 0:
            target.occupant.take_consequences(damage, 'damage', 'permanent')
        elif damage < 0:
            agent.take_consequences((-1) * damage, 'damage', 'temporary')
            
        elif damage <= 0:
            message("{0} {3} miss{1} {2}{4}".format(agent.name.capitalize(), '' if agent.is_player() else 'es', target.occupant.name, 
                        'critically' if damage < 0 else '', '.' if damage == 0 else ', hurting yourself!' if agent.is_player() else ', hurting itself!'
                                                ))
        if damage < 0:
            agent.take_consequences((-1) * damage, 'damage', 'permanent')
    if target.feature is not None:
        message("The {0} destroys {1}!".format(action.piece.name, target.feature.name), 'fiery')
        target.feature = None
    target.terrain.name = 'blackpowder-blasted ground' 
    target.terrain.shape = 'star'
    target.terrain.color = 'grave grey'
    target.terrain.flags = []
    return True

def trait_transfer(action, agent):
    """ Damage transfer, actually."""
    successes, advantages = resolve_task(agent['attitude'],agent['skill'], agent['luck'], agent['confidence'])
    successes += 1
    advantages += 1
    if successes > 0 or advantages > 0:
        message(action['msg'], action['msg color'])
        if successes > 0:
            agent.take_consequences(successes, action['from'], 'permanent')
            agent.take_consequences(successes, action['to'], 'permanent')
        if advantages > 0:
            agent.take_consequences(advantages, action['side effect'], 'temporary')
    else:
        message("You've dropped your {0} before you could use it.".format(action.piece.name))
    return True

def chain_lightning(action, agent):
    dmg_dice, current_range = action['damage'], action['range']
    x = agent.x
    y = agent.y
    while dmg_dice > 0:
        target = closest_agent(x, y, current_range)
        if target is None:
            infotip('No enemy in range.', 'failure')
            break
        else:
            dmg = _damage(dmg_dice)
            message('Electricity arcs to {0}. Damage due to electrocution is {1}.'.format(target.name, dmg), 'electric')
            target.take_damage(dmg)
            x, y = target.x, target.y
        dmg_dice -= 1

def confusing_melee(action, agent, target):
    target_health = target['health']
    melee(action, agent, target)
    if target_health > target['health'] and not target.has_flag('dead'):
        confusion_duration = (target_health - target['health']) / 4
        confusion_duration = confusion_duration if confusion_duration > 0 else 1 
        confusion_ai = Confusion(target['ai'], confusion_duration)
        target['ai'] = confusion_ai
        message('The eyes of the {0} look vacant, as he starts to stumble around! Confusion will last {1} turn{2}.'. \
                format(target.name, confusion_duration, 's' if confusion_duration > 1 else '')
                , 'confusion')    

def throw_grenade(action, agent, target):
    weapon = action.piece
    hit = coinflip()
    x, y = target.x, target.y
    msg = ''
    if not hit:
        if agent.is_player():
            msg += "Your aim is a little off."
        deviation = random_choice(eight_directions)
        x += deviation[0]
        y += deviation[1]
    damage = _damage(action['damage'])
    msg += ("{0} throw{1} {2}. It explodes, dealing {3} damage within blast radius ({4})!".format(
                agent.name.capitalize(), '' if agent.is_player() else 's',
                articled(weapon.name), damage, action['radius']))
    message(msg)
    tiles_in_blast = game.board.circle(x,y,action['radius'])
    for tile in tiles_in_blast:
        if tile.occupant is not None:
            tile.occupant.take_damage(damage)

def shoot_or_reload(action, agent, target = None):
    weapon = action.piece
    if weapon['loaded']:
        shoot(action, agent, target)
        weapon['loaded'] = False
        #weapon.name = weapon.name.replace('loaded', 'unloaded')
        action.requires_target = False
        #TODO: spend ammo
    else:
        action.requires_target = True
        weapon['loaded'] = True
        #weapon.name = weapon.name.replace('unloaded', 'loaded')
        message("{1} reload{2} {3} {0}.".format(weapon.name,
                                                    agent.name.capitalize(),'' if agent.is_player() else 's', 'your' if agent.is_player() else 'their'
                                                    ), weapon.color)

def _attack(action, agent, target):
    weapon = action.piece
    successes, advantages = resolve_task(agent['attitude'],agent['skill'], agent['luck'], -target['defense'])
    if successes > 0:
        if 'damage' in action:
            dmg_base = action['damage']
        else:
            dmg_base = 1 
        damage = successes * dmg_base
    elif successes < 0:
        damage = successes
        pass
    else:
        damage = 0
    return damage, advantages
    
def shoot(action, agent, target):
    weapon = action.piece
    if 'verb' not in action:
        action['verb'] = 'shoot'
    damage, advantages = _attack(action, agent, target)
    if damage > 0:
        message("{0} {6}{1} {4} at {2}{3} successfully!".format( #, causing {5} {7}!".format(
                agent.name.capitalize(),
                '' if agent.is_player() else 's',
                target.name,
                'rself' if target.is_player() and agent.is_player() else '',
                articled(weapon.name),
                damage,
                action['verb'],
                #action['success'][1]
                ))
        target.take_consequences(damage, action['success'][1], action['success'][0])
    elif damage <= 0:
        message("{0}{3} miss{1} {2}{4}".format(agent.name.capitalize(), '' if agent.is_player() else 'es', target.name, 
                        ' critically' if damage < 0 else '', '.' if damage == 0 else ', hurting yourself!' if agent.is_player() else ', hurting themself!'
                                                ))
        if damage < 0:
            agent.take_consequences((-1) * damage, 'damage', 'permanent')
    if advantages > 0:
        target.take_consequences(advantages, action['advantage'][1], action['advantage'][0])
    elif advantages < 0:
        agent.take_consequences((-1) * advantages, 'fatigue' if agent.is_player() else 'damage', 'permanent')
            
def melee(action, agent, target):
    weapon = action.piece
    if 'verb' not in action:
        action['verb'] = 'attack'
    damage, advantages = _attack(action, agent, target)
    if damage > 0:
        message("{0} {5}{1} {2} with {3} successfully!".format(#, causing {4} {6}!".format(                                                                                 
                agent.name.capitalize(),
                '' if agent.is_player() else 's',
                target.name,
                articled(weapon.name),
                damage,
                action['verb'],
                #action['success'][1]
                ))
        target.take_consequences(damage, action['success'][1], action['success'][0])
    elif damage <= 0:
        message("{0}{3} miss{1} {2}{4}".format(agent.name.capitalize(), '' if agent.is_player() else 'es', target.name, 
                        ' critically' if damage < 0 else '', '.' if damage == 0 else ', hurting yourself!' if agent.is_player() else ', hurting itself!'
                                                ))
        if damage < 0:
            agent.take_consequences((-1) * damage, 'damage', 'permanent')
    if advantages > 0:
        target.take_consequences(advantages, action['advantage'][1], action['advantage'][0])
    elif advantages < 0:
        agent.take_consequences((-1) * advantages, 'fatigue' if agent.is_player() else 'damage', 'permanent')
