'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

from guts.game_data import game
from guts.utils import *
from guts.messages import message

class AI(object):
    
    def __init__(self, previous_ai = None, countdown = None):
        self.previous_ai = previous_ai
        self.countdown = countdown
    
    def act(self, agent):
        """ A virtual method, so to speak."""
        raise Exception("Unimplemented AI method : " + self.__class__.__name__)

    def go_for_melee(self, agent, target):
        if distance(agent.x, agent.y, target.x, target.y) > 1:
            self.move_towards(agent, target)
        else:
            agent.melee(target)

    def go_for_ranged(self, agent, target):
        """ The agent tries to position itself to get a clean shot
            at the target or fires.
        """
        agent.shoot(target)

    def move_towards(self, agent, target):
        acceptable_directions = []
        for dpos in eight_directions:
            dx, dy = dpos
            x, y = agent.x + dx, agent.y + dy
            if not game.board[x,y].is_blocked():
                acceptable_directions.append((dx, dy, distance(x,y, target.x, target.y)))
        if len(acceptable_directions) == 0:
            agent.wait()
        else:
            dx, dy, _ = sorted(acceptable_directions,key=lambda direction: direction[2])[0]
            x, y = agent.x + dx, agent.y + dy
            agent.move_to(x,y)

    def restore_previous(self, agent):
        agent['ai'] = self.previous_ai
    
    def random_step(self, agent):
        x = agent.x + random_int(-1, 1)
        y = agent.y + random_int(-1, 1)
        if game.board[x,y].is_blocked():
            if not game.board[x,y].occupant == agent:
                message("{0} bumps into {1}.".format(agent.name, game.board[x,y].blocking_piece))
            else:
                message("{0} spins around in confusion.".format(agent.name))
        else:
            agent.move_to(x,y)

class RelentlessMelee(AI):
    
    def act(self, agent):
        if game.in_fov(agent.x, agent.y):
            self.go_for_melee(agent, game.player)
        else:
            agent.wait()

class StandardSoldier(AI):
    
    def act(self, agent):
        if game.in_fov(agent.x, agent.y):
            if distance(agent.x, agent.y, game.player.x, game.player.y) == 1:
                self.go_for_melee(agent, game.player)
            else:
                self.go_for_ranged(agent, game.player)
        else:
            agent.wait()

class BossCaster(AI):

    def act(self, agent):
        agent.wait()
 
class NightmareTyrant(AI):
    
    def transform(self, agent):
        agent.wait()
    
    def act(self, agent):
        if distance(agent.x, agent.y, game.player.x, game.player.y) == 1 and len(agent['forms']) > 0:
            #teleport away
            wrong_throne = (agent.x, agent.y)
            possible_thrones = [(throne[0], throne[1]) for throne in agent['thrones'] if throne[0]!= wrong_throne[0] or throne[1]!= wrong_throne[1]]
            selected_throne = random_choice(possible_thrones)
            message("The Nightmare Tyrant merges with its throne and emerges at another one!", 'dark nightmare')
            agent.move_to(selected_throne[0],selected_throne[1])
            agent['forms'].pop()
        elif game.in_fov(agent.x, agent.y):
            self.go_for_ranged(agent, game.player)
        else:
            agent.wait()
 
class Confusion(AI):
        
    def act(self, agent):
        if self.countdown == 0:
            self.restore_previous(agent)
            if not isinstance(agent['ai'], Confusion):
                message('The ' + agent.name + ' is no longer confused!', 'clarity')
        else:
            self.countdown -= 1
            self.random_step(agent) 

#__all__ = ['RelentlessMelee', 'StandardSoldier', 'Confusion']    