'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

from game_pieces import *
from game_tileboard import *
import libtcodpy as libtcod
from utils import *
from guts.game_piece_defs import *
from guts.game_quests import Objective
from guts.debug import debug
from guts.game_data import game

class Rect(object):
    #a rectangle on the map. used to characterize a room.
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.x_end = x + width + 1
        self.y_end = y + height + 1
        self.w = width
        self.h = height
    
    def center(self):
        center_x = (self.x + self.x_end) / 2
        center_y = (self.y + self.y_end) / 2
        return (center_x, center_y)
 
    def covers(self, x, y):
        return (x >= self.x and x < self.x_end) and (y >= self.y and y < self.y_end)
 
    def intersect(self, other):
        #returns true if this rectangle intersects with another one
        return (self.x  < other.x_end and self.x_end > other.x and
                self.y  < other.y_end and self.y_end > other.y)

    def __repr__(self):
        return str(self.__class__) + str((self.x,self.y,self.x_end,self.y_end))

def generate_board(width, height, player, mission):
    generator = BoardGenerator(mission, width, height)
    generator.generate()
    generator.place_player(player)
    return generator.board

class BoardGenerator(object):
    
    def __init__(self, mission, board_width, board_height):
        self.board = Board(board_width, board_height)
        self.mission = mission
        self.rooms = []
        self.default_floor = 'floor'
    
    def generate(self):
        self.build_board(self.mission)
        self.populate_board(self.mission)
    
    def spawn_posse(self, x, y, posse_member_type, posse_number, posse_member_name = None):
        acceptable_directions = []
        for direction in compass_directions:
            dx, dy = directions[direction]
            if not self.board[x + dx, y + dy].is_blocked():
                acceptable_directions.append(direction)
        bodyguard_spawns = random_combination(acceptable_directions, min(len(acceptable_directions), posse_number))
        for spawn_point in bodyguard_spawns:
            dx, dy = directions[spawn_point]
            self.board[x + dx, y + dy].occupant = new_agent(posse_member_type, posse_member_name)
    
    def put_doors_in(self, room):
        for x in range(room.x, room.x_end):
            if not self.board[x, room.y].is_blocked():
                self.board[x, room.y].feature = new_feature('wooden door')
            if not self.board[x, room.y_end].is_blocked():
                self.board[x, room.y_end].feature = new_feature('wooden door')
        for y in range(room.y, room.y_end):
            if not self.board[room.x, y].is_blocked():
                self.board[room.x, y].feature = new_feature('wooden door')
            if not self.board[room.x_end, y].is_blocked():
                self.board[room.x_end, y].feature = new_feature('wooden door')

    
    def create_room(self, room, terrain = None):
        if terrain is None:
            terrain = self.default_floor
        #go through the tiles in the rectangle and make them passable
        for x in range(room.x + 1, room.x_end):
            for y in range(room.y + 1, room.y_end):
                self.board[x,y].terrain = new_terrain(terrain)
        self.rooms.append(room)
        return room

    def create_h_tunnel(self, x1, x2, y, terrain = None):
        if terrain is None:
            terrain = self.default_floor
        for x in range(min(x1, x2), max(x1, x2) + 1):
            self.board[x,y].terrain = new_terrain(terrain)

    def create_v_tunnel(self, y1, y2, x, terrain = None):
        if terrain is None:
            terrain = self.default_floor
        #vertical tunnel
        for y in range(min(y1, y2), max(y1, y2) + 1):
            self.board[x,y].terrain = new_terrain(terrain)
    
    def prepare_boss_room(self, side, room_width, room_length, floor, wall):
        room_x_size = room_length if side in ['W','E'] else room_width
        room_y_size = room_length if side in ['N','S'] else room_width
        if side in ['E', 'W']:
            room_x = self.board.width - 3 if side == 'W' else 2 
            room_y = random_int(2, self.board.height - 3 - room_y_size)
        else:
            room_x = random_int(2, self.board.width - 3 - room_x_size)
            room_y = self.board.height - 3 if side == 'N' else 2 
        x_sign = 1 if side == 'W' else -1
        y_sign = 1 if side == 'N' else -1
        room = Rect(room_x - room_x_size if side == 'W' else room_x, room_y - room_y_size if side == 'N' else room_y, room_x_size, room_y_size)
        for tile in self.board.rectangle(room.x, room.y, room_x_size, room_y_size):
            tile.terrain = new_terrain(wall)
        for tile in self.board.rectangle(room.x + 1, room.y + 1, room_x_size - 2, room_y_size - 2, filled = True):
            tile.terrain = new_terrain(floor)
        return room, x_sign, y_sign, room_x_size, room_y_size
    
    def create_witch_hut(self, side):
        room_width = 11
        room_length = 7
        room, x_sign, y_sign, room_x_size, room_y_size = self.prepare_boss_room(side, room_width, room_length,
                         floor = 'wooden floor', wall = 'wooden wall')
        door = 'wooden door'
        if side in ['E', 'W']:
            x = room.x if side == 'W' else room.x + room_x_size
            y = room.y + room_y_size/2
            xx = x + 3 * x_sign
            yy = y + 2
        else: # side in [N, W]:
            x = room.x + room_x_size/2
            y = room.y if side == 'N' else room.y + room_y_size
            xx = x + 2
            yy = y + 3 * y_sign
        self.board[x, y].terrain = new_terrain('wooden floor')
        self.board[x, y].feature = new_feature('wooden door')
        self.board[xx, yy].occupant = new_agent('the Witch')
        self.spawn_posse(xx, yy, 'deerman', 1)
        self.spawn_posse(xx, yy, 'boarman', 1)
        self.spawn_posse(xx, yy, 'wolfman', 1)
        return room
    
    def create_crime_lords_office(self, side):
        room_width = 6
        room_length = 4
        room, x_sign, y_sign, room_x_size, room_y_size = self.prepare_boss_room(side, room_width, room_length,
                         floor = 'wooden floor', wall = 'stone wall')
        door = 'wooden door'
        if side in ['E', 'W']:
            x = room.x if side == 'W' else room.x + room_x_size
            y = room.y + room_y_size/2
            for i in range(3):
                self.board[x + 2 * x_sign, y + i - 1].feature = new_feature('antique desk')
                if i == 1:
                    mod = 1 if side == 'W' else -1
                    self.board[x + 2 * x_sign + mod, y + i - 1].occupant = new_agent('the Crime Lord')
            self.board[x + 1 * x_sign, room.y + 1].occupant = new_agent('thug')
            self.board[x + 1 * x_sign, room.y + room_y_size - 1].occupant = new_agent('thug')
        else: # side in [N, W]:
            x = room.x + room_x_size/2
            y = room.y if side == 'N' else room.y + room_y_size
            for i in range(3):
                self.board[x + i - 1,y + 2 * y_sign].feature = new_feature('antique desk')
                if i == 1:
                    mod = 1 if side == 'N' else -1
                    self.board[x + i - 1 ,y + 2 * y_sign + mod].occupant = new_agent('the Crime Lord')
            self.board[room.x + 1,y + 1 * y_sign].occupant = new_agent('thug')
            self.board[room.x + room_x_size - 1,y + 1 * y_sign].occupant = new_agent('thug')
        self.board[x, y].terrain = new_terrain('wooden floor')
        self.board[x, y].feature = new_feature('wooden door')

        return room
    
    def create_pipers_cave(self, side):
        room_width = 4
        room_length = 8
        room, x_sign, y_sign, room_x_size, room_y_size = self.prepare_boss_room(side, room_width, room_length,
                         floor = 'dirt', wall = 'slag heap of trash')
        door = 'wooden door'
        trash_piles = []
        rat_enemies = []
        door = None
        lair = None
        piper = None
        if side in ['E', 'W']:
            x = room.x if side == 'W' else room.x + room_x_size
            y = room.y + room_y_size/2
            door = (x,y)
            mod = random_choice([-1,1])
            piper = (x + 7, y + mod)
            lair = (x + 7, y - mod)
            for i in range(3):
                rat_enemies.append((x + 2 * (i+1),y))
                trash_piles.append((x + 2 * (i+1),y + mod * ((-1) ** i)))
        else: # side in [N, S]:
            x = room.x + room_x_size/2
            y = room.y if side == 'N' else room.y + room_y_size
            door = (x,y)
            mod = random_choice([-1,1])
            piper = (x + mod, y + 7)
            lair = (x - mod, y + 7)
            for i in range(3):
                rat_enemies.append((x,y + 2 * (i+1)))
                trash_piles.append((x + mod * ((-1) ** i),y + 2 * (i+1)))
        self.board[door[0],door[1]].terrain = new_terrain('dirt')
        self.board[door[0],door[1]].feature = new_feature('wooden door')
        self.board[lair[0],lair[1]].feature = new_feature('filthy mattress')
        self.board[piper[0],piper[1]].occupant = new_agent('the Piper')
        for (x,y) in trash_piles:
            self.board[x,y].feature = new_feature('pile of trash')
        for (x,y) in rat_enemies:
            self.board[x,y].occupant = new_agent(random_choice(['rabid ratling', 'ratling scavenger', 'ratling warrior', 'rat hound']))
        return room
    
    def create_heart_of_nightmares(self, side):
        room_width = self.board.width - 3 if side in ['E', 'W'] else self.board.height - 3
        room_length = 4
        room, x_sign, y_sign, room_x_size, room_y_size = self.prepare_boss_room(side, room_width, room_length,
                         floor = 'golden floor', wall = 'golden wall')
        #TODO: properly initialize the nightmare tyrant
        throne_gap = room_width/4
        nightmare_tyrant = new_agent('the Nightmare Tyrant')
        throne_tiles = []
        if side in ['E', 'W']:
            x = room.x if side == 'W' else room.x + room_x_size
            y = room.y + 1
            while y < room_width + room.y:
                self.board[x, y].terrain = new_terrain('nightmarish ground')
                y += 2
            y = room.y
            
            for i in range(3):
                throne_tiles.append((x + 2 * x_sign, y + throne_gap * (i+1)))
                self.board[throne_tiles[i][0], throne_tiles[i][1]].terrain = new_terrain('throne of nightmares')
        else: # side in [N, W]:
            y = room.y if side == 'N' else room.y + room_y_size
            x = room.x + 1
            while x < room_width + room.x:
                self.board[x, y].terrain = new_terrain('nightmarish ground')
                x += 2
            x = room.x
            for i in range(3):
                throne_tiles.append((x + throne_gap * (i+1),y + 2 * y_sign))
                self.board[throne_tiles[i][0], throne_tiles[i][1]].terrain = new_terrain('throne of nightmares')
        initial_throne = random_choice(throne_tiles)
        self.board[initial_throne[0],initial_throne[1]].occupant = nightmare_tyrant 
        nightmare_tyrant['thrones'] = throne_tiles
        forms = []
        for quest in game.quests:
            if not quest.is_completed():
                nightmare_tyrant['health'] += 10
                forms.append(quest.name)
        nightmare_tyrant['forms'] = forms
        return room

    def create_necromancer_crypt(self, side):
        room_width = 9
        room_length = 8
        room, x_sign, y_sign, room_x_size, room_y_size = self.prepare_boss_room(side, room_width, room_length,
                         floor = 'crypt floor', wall = 'crypt wall')
        geists = random_permutation(['ghost', 'poltergeist'])
        if side in ['E', 'W']:
            xx = x = room.x if side == 'W' else room.x + room_x_size
            y = room.y + room_y_size/2
            yy = y + 1
            for x_mod in range(1,3):
                for y_mod in range(1,3):
                    self.board[x + 2 * x_mod * x_sign, y - y_mod].feature = new_feature('sarcophagus')
                    self.board[x + 2 * x_mod * x_sign, y +1 + y_mod].feature = new_feature('sarcophagus')
            mod = -1
            if coinflip():
                self.board[x + 6 * x_sign, y - 1].occupant = new_agent('zombie') if coinflip() else new_agent('skeleton')  
                mod = 0
            else:
                self.board[x + 6 * x_sign, y + 2].occupant = new_agent('zombie') if coinflip() else new_agent('skeleton')
            for i in range(3):
                self.board[x + 6 * x_sign, y + i + mod].feature = new_feature('ancient desk')
                if i == 1:
                    self.board[x + 7 * x_sign, y + i + mod].occupant = new_agent('the Necromancer')
            self.board[x + 7 * x_sign, room.y + 1].occupant = new_agent(geists[0])
            self.board[x + 7 * x_sign, room.y + room_y_size - 1].occupant = new_agent(geists[1])
        else: # side in [N, W]:
            x = room.x + room_x_size/2
            yy = y = room.y if side == 'N' else room.y + room_y_size
            xx = x + 1
            for x_mod in range(1,3):
                for y_mod in range(1,3):
                    self.board[x - x_mod, y + 2 * y_mod * y_sign].feature = new_feature('sarcophagus')
                    self.board[x + 1 + x_mod, y + 2 * y_mod * y_sign].feature = new_feature('sarcophagus')
            mod = -1
            if coinflip():
                self.board[x - 1, y + 6 * y_sign].occupant = new_agent('zombie') if coinflip() else new_agent('skeleton')  
                mod = 0
            else:
                self.board[x + 2, y + 6 * y_sign].occupant = new_agent('zombie') if coinflip() else new_agent('skeleton')
            for i in range(3):
                self.board[x + i + mod,y + 6 * y_sign].feature = new_feature('ancient desk')
                if i == 1:
                    self.board[x + i + mod,y + 7 * y_sign].occupant = new_agent('the Necromancer')
            self.board[room.x + 1,y + 7 * y_sign].occupant = new_agent(geists[0])
            self.board[room.x + room_x_size - 1,y + 7 * y_sign].occupant = new_agent(geists[1])
        self.board[x, y].terrain = new_terrain('crypt floor')
        self.board[xx, yy].terrain = new_terrain('crypt floor')
        self.board[x, y].feature = new_feature('ancient door')
        self.board[xx, yy].feature = new_feature('ancient door') 
        return room
        
    def create_player_room(self, side):
        room_size = 1
        side_coords = {'W' : (1, self.board.width/2),
                       'N' : (self.board.width/2, 1),
                       'E' : (self.board.width - room_size - 1, self.board.width/2),
                       'S' : (self.board.width/2, self.board.height - room_size - 1)
                       }
        return Rect(side_coords[side][0], side_coords[side][1], room_size, room_size)
    
    boss_room_by_quest = {
                         'Haunted Graveyard' : create_necromancer_crypt, 
                         'Cursed Forest' : create_witch_hut,
                         'Black Market' : create_crime_lords_office,
                         'Rat-infested Landfill' : create_pipers_cave,
                         'Nightmare World' : create_heart_of_nightmares 
                         }
    
    board_by_quest = {'Haunted Graveyard' : {'outer wall': 'graveyard fence',
                                             'floor': 'graveyard soil',
                                             'special floor': 'toppled tombstone',
                                             'common feature': 'tombstone',
                                             'rare feature': 'withered tree',
                                             'rare floor': 'memorial monument',
                                            }, 
                      'Cursed Forest' : {    'outer wall': 'thorny thicket',
                                             'floor': 'grass',
                                             'special floor': 'small stream',
                                             'rare floor': 'thorny thicket',
                                             'common feature': 'shrubbery',
                                             'rare feature': 'gnarled tree',
                                            }, 
                      'Black Market' : {     'outer wall': 'stone wall',
                                             'floor': 'dirt',
                                             'special floor': 'cobblestone road',
                                             'rare floor': 'shelf full of illicit wares',
                                             'common feature': 'stall',
                                             'rare feature': 'cart',
                                            }, 
                      'Rat-infested Landfill' : {'outer wall': 'slag heap of trash',
                                                 'floor': 'dirt',
                                                 'special floor': 'muddy puddle',
                                                 'rare floor': 'mound of trash',
                                                 'common feature': 'pile of trash',
                                                 'rare feature': 'heap of trash',
                                            }, 
                      'Nightmare World' : {  'outer wall': 'dreaming townsperson',
                                                'floor': 'nightmarish ground',
                                                'special floor': 'wicked weed',
                                                'rare floor': 'swirling horror',
                                                'common feature': 'nightmarish tree',
                                                'rare feature': 'horrific statue',
                                            }, 
                      }

    def build_board(self, final_stage = False):
        """
        for x in range(0, self.board.width):
            self.board[x, 0].terrain = new_terrain('graveyard fence')
            self.board[x, self.board.height].terrain = new_terrain('crypt floor')
        for y in range(0, crypt_y + crypt_y_size * y_inc, y_inc):
            self.board[crypt_x, y].terrain = new_terrain('crypt floor')
            self.board[crypt_x + crypt_x_size * x_inc, y].terrain = new_terrain('crypt floor')
    
        for tile in self.board.area(0,0,0, self.board.width, 0, self.board.height, \
                            excluded = self.board.area(1,1,self.board.width - 1, 0, self.board.height - 1)):
            tile.terrain = new_terrain('graveyard fence'
        """
        for tile in self.board.rectangle(0,0,self.board.width -1, self.board.height -1):
            tile.terrain = new_terrain(self.board_by_quest[self.mission]['outer wall'])
        entrance_side = random_choice(cardinal_directions)
        
        #prepare the necromancer's crypt
        main_room = self.boss_room_by_quest[self.mission](self, entrance_side)
        
        for x in range(1, self.board.width - 1):
            for y in range(1, self.board.height - 1):
                if not main_room.covers(x,y):
                    if coinflip():
                        self.board[x,y].terrain = new_terrain(self.board_by_quest[self.mission]['floor'])
                    else:
                        if coinflip():
                            if random_pos(4) == 4:
                                self.board[x,y].terrain = new_terrain(self.board_by_quest[self.mission]['rare floor'])
                            else:
                                self.board[x,y].terrain = new_terrain(self.board_by_quest[self.mission]['special floor'])
                        else:
                            self.board[x,y].terrain = new_terrain(self.board_by_quest[self.mission]['floor'])
                            if random_pos(4) == 4:
                                self.board[x,y].feature = new_feature(self.board_by_quest[self.mission]['rare feature'])
                            else:
                                self.board[x,y].feature = new_feature(self.board_by_quest[self.mission]['common feature'])
        """                
        for x in range(0, self.board.width):
            for y in range(0, self.board.height):
                if self.board[x,y].terrain is None:
                    print(x,y)
        """
        #self.put_doors_in(main_room)        
        self.rooms = [self.create_player_room(entrance_side), main_room]
        
    def occupy_room(self, room, agent_type):
        available_tiles = []
        for tile in self.board.rectangle(room.x, room.y, room.w, room.h):
            if not tile.is_blocked():
                available_tiles.append(tile)
        if available_tiles:
            tile = random_choice(available_tiles)
            tile.occupant = new_agent(agent_type)
            return True
        else:
            return False

    monster_populations_by_quest = {
        'Haunted Graveyard': ['zombie'] * 5 + ['skeleton'] * 5 + ['poltergeist'] * 3 + ['dread bat'] * 2 + ['ghost'] + \
                ['zombie' if coinflip() else 'skeleton'] + ['zombie' if coinflip() else 'skeleton'] + \
                ['dread bat' if random_pos(3) == 1 else 'ghost' if coinflip() else 'poltergeist'] + \
                ['dread bat' if random_pos(3) == 1 else 'skeleton' if coinflip() else 'zombie'],
        'Cursed Forest' : ['deerman'] * 6 + ['boarman'] * 6 + ['wolfman'] * 6 + ['wasp swarm'] * 2,
        'Rat-infested Landfill' : ['ratling warrior'] * 5 + ['ratling scavenger'] * 5 + ['rabid ratling'] * 5 + ['rat hound'] * 5,
        'Black Market' : ['thug'] * 5 + ['fence'] * 5 + ['thief'] * 5 + ['guard dog'] * 5,
        'Nightmare World' : ['the Unknown Fear'] * 6
        }
    
    nightmare_monsters = {
                         'Haunted Graveyard' : 'the Fear After',
                         'Cursed Forest' : 'the Fear Outside',
                         'Black Market' : 'the Fear Inside',
                         'Rat-infested Landfill' : 'the Fear Below' 
                         }
        
    def populate_board(self, final_stage = False):
        monster_population = self.monster_populations_by_quest[self.mission] 
        if self.mission == 'Nightmare World':
            for quest in game.quests:
                if not quest.is_completed() and not quest.name == 'Nightmare World':
                    monster_population += [self.nightmare_monsters[quest.name]] * 4
        #print(monster_population)
        monster_population = random_permutation(monster_population)
        room_width = self.board.width / 5
        room_height = self.board.width / 5
        rooms = [[Rect(i * room_width, j * room_height, room_width, room_height) for i in range (5)] for j in range(5)]
        for room_row in random_permutation(rooms):
            for room in random_permutation(room_row):
                if not room.intersect(self.rooms[1]) and monster_population:
                    agent_type = monster_population.pop()
                    if not self.occupy_room(room, agent_type):
                        monster_population.append(agent_type)
        pass
        
    def place_player(self, player):
        first_room = self.rooms[0]
        (player_x, player_y) = first_room.x + min(first_room.w - 1,2), first_room.y + min(first_room.h - 1,2)
        self.board[player_x, player_y].occupant = player
