'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

from guts.debug import debug
from guts.game_data import game
from guts.messages import message, articled, infotip
from guts.utils import coinflip, random_pos
from guts.game_piece import Piece
from guts.game_description import trait_level_description, duration_description
from guts.task_resolution import traits, trait_losses

item_action_types = ['melee', 'ranged', 'deployable', 'activated', 'consumable']

class Agent(Piece):
    
    def __init__(self, *args, **kwargs):
        Piece.__init__(self, *args, **kwargs)
        self['role'] = 'unknown'
        self.inventory = []
        for flag in item_action_types:
            self[flag + ' item'] = None

    def leave(self):
        game.board[self.x, self.y].occupant = None
        
    def _arrive_on(self, tile):
        tile.occupant = self
    
    def change_role(self, role):
        self['role'] = role

    def is_player(self):
        return self['role'] == 'player'

    def take_consequences(self, amount, consequence, duration):
        if self.has_flag('dead'):
            return None
        assert(amount >= 0)
        trait_affected = consequence if consequence in traits else trait_losses[consequence]
        max_trait = self['max ' + trait_affected] if trait_affected in ['health', 'stamina'] else 3
        min_trait = 0 if trait_affected in ['health', 'stamina'] else -3
        
        original_trait_value = self[trait_affected]
        if consequence in trait_losses:
            self[trait_affected] = max(min_trait, original_trait_value - amount)
        else:
            self[trait_affected] = min(max_trait, original_trait_value + amount)
        amount = (original_trait_value - self[trait_affected]) * (1 if consequence in trait_losses else -1)
        #TODO: spillover -> stamina -> health
        
        if duration != 'permanent':
            assert duration in ['temporary', 'tactical']
            self[duration + ' ' + trait_affected] += amount
        if trait_affected is not None:
            message("{0}{1} {2} {5}{6}{3}s by {4}.".format(self.name.capitalize(), 'r' if self.is_player() else "'s",
                                                     trait_affected, "decrease" if consequence in trait_losses else "increase",
                                                     amount if amount >= 0 else -amount, duration_description(duration),
                                                     ' ' if len(duration_description(duration)) > 0 else ''), consequence)
        if self['health'] == 0:
            self.death()

    def _refresh(self, duration):
        refreshed = False
        for trait in traits:
            if self[duration + ' ' + trait] != 0:
                self.take_consequences(self[duration + ' ' + trait],
                                       trait if self[duration + ' ' + trait] >0 else trait_gains[trait], 'permanent')
                self[duration + ' ' + trait] = 0
                refreshed = True
        if self['ranged item'] is not None:
            if not self['ranged item']['loaded']:
                self['ranged item'].use(self, 'ranged')
                refreshed = True
        return refreshed
        
    def refresh(self, full = False):
        #tactical refresh
        refreshed = self._refresh('tactical')
        if full:
            refreshed = refreshed or self._refresh('temporary')
        return refreshed

    def take_damage(self, damage):
        damaged_trait = None
        if damage > 0:
            if self.has('shields'):
                if damage > self['shields']:
                    damage -= self['shields']
                    self['shields'] = 0
                else:
                    self['shields'] -= damage
                    damage = 0
                    damaged_trait = 'shields'
            if self.has('armor'):
                damage = 0 if self['armor'] >= damage else damage - self['armor']
            if damage > 0:
                self['health'] -= damage
                if self['health'] <= 0:
                    self.death()
                    damaged_trait = None
                else:
                    damaged_trait = 'health'
        if damaged_trait is not None and not self.is_player():
            trd = trait_level_description(damaged_trait, self)
            message("{0} is {1}.".format(self.name.capitalize(), trd), trd)

    def death(self):
        message(self.name.capitalize() + " dies!", 'frag')
        if not self.is_player():
            game.note_death(self)
        #for item in [item for item in self.inventory]:
        #item.drop(self)
        self.become_corpse()
        self.set_flag('dead')
        self.ai = None
        self.leave()
    
    def become_corpse(self):
        #TODO: corpses! 
        '''monster.char = 'corpse'
        monster.color = 'dark_red'
        monster.name = 'remains of ' + monster.name
        '''
        pass

    def act(self):
        self['ai'].act(self)

    def wait(self):
        pass

    def drop(self, item):
        # remove from agent's inventory and put it on the tile the agent is standing on.
        self.inventory.remove(self)
        game.board[self.x, self.y].put(self)
        if self.is_player():
            message('You dropped a ' + self.name + '.')

    def pick_up(self, item):
        """ TODO: inventory limits or carrying capacity or both
        if agent.has_stat('carrying capacity'): 
        if len(Game().inventory) >= 26:
            infotip('Your inventory is full, cannot pick up ' + self.owner.name + '.', 'warning')
        else:
        """
        game.board[item.x, item.y].remove(item)
        self.receive(item)
        if self.is_player:
            message('You picked up {0}!'.format(articled(item.name)), 'positive')

    def receive(self, item):
        if item.has_flag('stacks') and item['type'] in [inv_item['type'] for inv_item in self.inventory]:
            item_stack = filter(lambda f_item : f_item['type'] == item['type'], self.inventory)[0]
            if not item_stack.has('stack'):
                item_stack['stack'] = 1
            item_stack['stack'] += item['stack'] if item.has('stack') else 1
        else:
            self.inventory.append(item)

    def equip(self, item, silently = True):
        message_fired = silently
        items_to_remove = []
        for action_type in item_action_types:
            if item.has(action_type):
                items_to_remove.append(self[action_type + ' item'])
                if self[action_type + ' item'] != item:
                    self[action_type + ' item'] = item
                    """ There are no items in Nightmare Tyrant to pick up, no point in telling the player about equipping them 
                    if not message_fired and self.is_player() or game.in_fov(self.x, self.y):
                        message("{0} equip{1} {2}.".format(str.capitalize(self.name), '' if self.is_player() else 's', articled(item.name)), 'positive')
                        message_fired = True
                    """
        for action_type in item_action_types:
            if self[action_type + ' item'] in filter(None, items_to_remove):
                self[action_type + ' item'] = None
        """ see above re: equipping items
        for equipped_item in filter(None, items_to_remove):
            message("{0} unequip{1} {2}.".format(str.capitalize(self.name), '' if self.is_player() else 's', articled(equipped_item.name)), 'negative')
        """
        #TODO: apply passive effects

    def melee(self, target = None):
        if self['melee item'] is None:
            debug.log(self.name + " has no melee item, but tries to melee something.")
            if self.is_player():
                infotip("You can't touch the " + target.name, 'hot pink')
        else:
            self['melee item'].use(self, 'melee', target)

    def shoot(self, target = None):
        self['ranged item'].use(self, 'ranged', target)
        
class Player(Agent):

    def __init__(self, *args, **kwargs):
        Agent.__init__(self, *args, **kwargs)
        self['role'] = 'player'
        
    def _arrive_on(self, tile):
        
        Agent._arrive_on(self, tile)
        
    def death(self):
        message('You died!','blood red')
        game.state = 'dead'
        self.set_flag('dead')
        #for added effect, transform into a corpse!
        #TODO: actual dropping of a corpse in case of teamplay - probably just needs a switch in agent death()
        self.shape = 'corpse'
        self.color = 'blood red'
        infotip('Press <SPACE> or <ESC> to go back to main menu', 'important')

    def shoot(self, target):
        self['ranged item'].use(self, target)
        
