'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

import shelve, dbhash
import libtcodpy as libtcod
from guts.config import FOV_ALGO, FOV_LIGHT_WALLS
from collections import defaultdict

class GameData():
    """ Game data stores all the data that is generated for the actual game.
        Part of the data is primary - that is, it is generated procedurally 
        at the beginning of the game or by player's choice or actions and
        therefore needs to be stored in a file when saving a game.
        Then, there's secondary data, that can be regenerated from the primary
        data, by simple operations (done here) or using external methods 
        - the latter data must be passed to the object of this class 
        as parameters to ensure that this module is the root dependency
        of the project.
        
    """  
    
    def __init__(self):
        """ Saving and loading data needs to make sure it doesn't store
            any object twice independently (e.g. player as a piece on 
            the board and then self.player handle), because shelve will
            NOT keep these connected. However, some pieces of the data
            might be annoying to recreate procedurally, so we can index
            them for ease of use.  
        """
        self.new()
        
    def new(self):
        #Primary data:
        self._board = None
        self.msgs = []
        self.msg_archive = []
        self.quests = []
        self.kill_list = defaultdict(int)
        self.time_limited_turns = False
        self.turncount = 0
        self.state = None
        #Secondary data - for storing and loading via index
        self.player = None
        self.current_quest = None
        #Secondary data - for procedural recreation
        self.fov_recompute = None
        self.fov_map = None
        #fleeting data - no need to store it
        self._infotips = []
        self._infotips_viewed = False
        
    def note_death(self, victim):
        self.kill_list[victim['name']] += 1
    
    def add_infotip(self, infotip, color):
        if self._infotips_viewed:
            self._infotips = []
            self._infotips_viewed = False
        self._infotips.append((infotip, color))
    
    def _get_infotips(self):
        self._infotips_viewed = True
        return self._infotips
    infotips = property(_get_infotips)
        
    def _get_pieces(self):
        pieces = []
        for tile in self.board.tiles:
            pieces.extend(tile.pieces)
        return pieces
    pieces = property(_get_pieces)
    
    def _get_agents(self):
        agents = []
        for tile in self.board.tiles:
            if tile.occupant is not None:
                agents.append(tile.occupant)
        return agents
    agents = property(_get_agents)
    
    def _get_board(self):
        return self._board
    def _set_board(self, board):
        self.fov_recompute = None
        self.fov_map = None
        self._pieces = None
        self._agents = None
        self._board = board
    board = property(_get_board, _set_board)
    
    def set_fov_at(self, x, y):
        libtcod.map_set_properties(self.fov_map, x, y, not self.board[x,y].blocks_sight, not self.board[x,y].is_blocked)
      
    def in_fov(self, x, y):
        #TODO: isn't it better to use game.board[x,y].visible?
        return libtcod.map_is_in_fov(self.fov_map, x, y) if game.state.find('briefing') < 0 else False
    
    def _get_tiles_in_fov(self):
        tiles = []
        for x in xrange(self.board.width):
            for y in xrange(self.board.height):
                if self.in_fov(x, y):
                    tiles.append(self.board[x,y])
        return tiles
    tiles_in_fov = property(_get_tiles_in_fov)
    
    def _get_agents_in_fov(self):
        agents = []
        for tile in self.tiles_in_fov:
            if tile.occupant is not None:
                agents.append(tile.occupant)
        return agents
    agents_in_fov = property(_get_agents_in_fov)
    
    def fov_check(self):
        #recompute FOV if needed (the player moved or something)
        if self.fov_recompute:
            libtcod.map_compute_fov(self.fov_map, self.player.x, self.player.y, self.player['sight'], FOV_LIGHT_WALLS, FOV_ALGO)
            self.fov_recompute = False
            return True
        return False
        
    def save(self, savegame):
        """ If the same object is shelved two times (e.g. by itself and as a part of
            a collection, weird things happen. Make sure it doesn't happen.
        """
        #open a new empty shelve (possibly overwriting an old one) to write the game data
        savefile = shelve.open(savegame, 'n')
        #primary data
        savefile['board'] = self.board
        savefile['quests'] = self.quests
        savefile['messages'] = self.msgs
        savefile['message archive'] = self.msg_archive
        savefile['game state'] = self.state
        savefile['time limited turns'] = self.time_limited_turns
        savefile['turncount'] = self.turncount
        savefile['kill list'] = self.kill_list
        #secondary data 
        if self.current_quest is not None:
            savefile['current quest'] = self.current_quest.name
        savefile['player position'] = (self.player.x, self.player.y)
        savefile['player'] = self.player #between visiting maps, the player might exist while the board doesn't, this handles such cases
        savefile.close()
        
    def load(self, savegame):
        self.new()
        #open the previously saved shelve and load the game data
        savefile = shelve.open(savegame, 'r')
        #primary data
        self.board = savefile['board'] 
        self.quests = savefile['quests']
        self.msgs = savefile['messages'] 
        self.msg_archive = savefile['message archive'] 
        self.state = savefile['game state'] 
        self.time_limited_turns = savefile['time limited turns']
        self.turncount = savefile['turncount']
        self.kill_list = savefile['kill list'] 
        #secondary data
        self.current_quest = [quest for quest in self.quests if quest.name == savefile['current quest']][0] if 'current quest' in savefile else None
        if self.board is not None: 
            player_x, player_y = savefile['player position']
            self.player = game.board[player_x, player_y].occupant 
        else:
            self.player = savefile['player']
        savefile.close()
        self.prepare_secondary_data()
    
    def prepare_secondary_data(self):
        #TODO: regenerate
        pass 

game = GameData()