'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

from guts.game_data import game
from guts.messages import infotip, message
from guts.debug import debug

class Objective(object):
    
    INTERACT = 'INTERACT'
    KILL = 'KILL'
    ARRIVE = 'ARRIVE'
    FETCH = 'FETCH'
    DELIVER = 'DELIVER'
    
    def __init__(self, id, description_todo, description_done, kind, target, prereqs = [], repeats = 1, optional = False):
        self.id = id
        self._description_todo = description_todo
        self._description_done = description_done 
        self.prereqs = prereqs
        self.kind = kind
        self.target = target
        self.repeats = repeats
        self.accomplished = 0
        self.optional = optional
        self.available = not prereqs
    
    def _get_description(self):
        if self.is_fulfilled():
            return self._description_done
        else:
            progress = ""
            if self.repeats > 1:
                progress = " {0}/{1}".format(self.accomplished, self.repeats)
            desc = "{0}{1}".format(self._description_todo, progress)
            return desc
    description = property(_get_description)
    
    def is_optional(self):
        return self.optional
    
    def is_available(self):
        return self.available

    def is_fulfilled(self):
        return self.accomplished >= self.repeats    

    def _get_status(self):
        if self.is_fulfilled():
            return 'fulfilled'
        if self.is_available():
            return 'available'
        else:
            return 'unavailable'
    status = property(_get_status)
    
    def check_progress(self):
        accomplishment_so_far = self.accomplished
        progress = False
        if self.kind == self.__class__.KILL:
            self.accomplished = game.kill_list[self.target]
            progress = self.accomplished > accomplishment_so_far 
        elif self.kind == self.__class__.ARRIVE:
            pass
        elif self.kind == self.__class__.FETCH:
            pass
        elif self.kind == self.__class__.DELIVER:
            pass            
        elif self.kind == self.__class__.INTERACT:
            pass            
        else:
            debug.log("Unknown kind of objective: {0}. Objective id: {1}".format(self.kind, self.id))
        if progress:
            message("{0} {1}".format("You have fulfilled an objective: " if self.is_fulfilled() else "You make progress on objective: ",
                                     self.description),'completed quest' if self.is_fulfilled() else 'tracked quest')
        return progress
    
    def __repr__(self):
        return ','.join([str(self.__class__),self.id, self.description, str(self.prereqs), str(self.kind), str(self.repeats), str(self.accomplished), str(self.optional)])

class Quest(object):
    '''
    classdocs
    '''

    UNAVAILABLE = 'unavailable'
    AVAILABLE = 'available'
    CURRENT = 'current'
    COMPLETED = 'completed'

    def __init__(self, name, description, status = AVAILABLE):
        '''
        * name to refer to the quest by
        * description to display in the HUD/quest selection info
        * objectives to complete during the quest (in a dictionary)
        * status - one of the enumerated states quest can have 
        '''
        self.name = name
        self.description = description
        self.status = status
        self.objectives = {}
    
    def is_completed(self):
        return self.status == self.__class__.COMPLETED

    def is_available(self):
        return self.status == self.__class__.AVAILABLE

    def is_unavailable(self):
        return self.status == self.__class__.UNAVAILABLE
    
    def add_goal(self, objective):
        self.objectives[objective.id] = objective
    
    def get_goal(self, objective_id):
        return self.objectives[objective_id]
    
        
    def complete(self):
        """ This is a proper place to put in hooks for quest manipulation.
            or is it?
            TODO: handling of quest trees (but for now the idea is to make 100% non-linear games)
        """
        self.status = self.__class__.COMPLETED
        
    def is_finished(self):
        for objective in self.objectives.values():
            if not objective.is_optional() and not objective.is_fulfilled():
                return False
        return True
    
    def update_availability_of_objectives(self):
        for objective in self.objectives.values():
            if objective.prereqs and not objective.is_available():
                make_available = True
                for objective_id in objective.prereqs:
                    if not self.objectives[objective_id].is_fulfilled():
                        make_available = False
                        break
                objective.available = make_available

    def accomplish(self, objective_id):
        self.objectives[objective_id].accomplished += 1
        self.update_availability_of_objectives()
        if self.is_finished():
            self.complete()
                
    def progress_objectives(self):
        progress = False
        for objective in [objective for objective in self.objectives.values() if objective.status == 'available']:
            progress = progress or objective.check_progress()
        if progress:
            self.update_availability_of_objectives()
            if self.is_finished():
                self.complete()
