'''
This code is part of NightmareTyrant; Copyright (C) 2013 Piotr 'ZasVid' Sikora; see NightmareTyrant.py for full notice.
'''

import os
import libtcodpy as libtcod
from guts.messages import message, a_or_an, articled, infotip
from guts.debug import debug
from guts.game_data import game
from guts.game_piece_defs import new_player
from guts.board_generation import generate_board
from guts.config import INFO_WIDTH
from guts.utils import compass_directions, directions, distance
from guts.player_input import input_manager
from guts.actions.combat import closest_agent
from guts.game_quests import Quest, Objective
from guts.game_goals_defs import prepare_quest, prepare_campaign


class GamePlay(object):
    '''
    This class handles playing the game.
    '''

    def __init__(self, ui, achievements):
        self.ui = ui
        self.achievements = achievements
        self.savefile = 'savegame'

    def new(self):
        """ New game is the place to setup everything for a new game to start, including:
            * reset of game data to default values 
            * ask the player for input for initialisation variables (e.g., character creation)
             
        """
        self.ui.clear()
        self.ui.msgbox('Welcome hero! The fate of this town rests in your hands. Can you save it from the approaching nightmare? Face ' \
                       'the nightmares anytime you want, but the more unresolved dangers lurk around the town, the stronger the fears ' \
                       'of its townspeople and the nightmare harder to conquer. However, you risk your life anytime you go out there to ' \
                       'face the frightening denizens of the towns most dangerous places.', 'light nightmare')
        game.new()
        #create object representing the player
        game.player = new_player()
        game.quests = prepare_campaign('Nightmare Tyrant')
        
        game.state = 'briefing'


    def load(self, savefile):
        self.ui.clear()
        game.load(savefile)
        
    def text_play(self):
        #display quests to choose from
        self.ui.clear()
        quit_game = False
        """ Debriefing stage """
        if game.state == 'debriefing' and game.player['stamina'] >= 0:
            """ Strategical refresh - after a quest is over remove all temporary effects
                from the player character, restock equipment etc. 
            """
            message('You rest and recover from the less serious ailments inflicted to you during your last quest.', 'completed quest')
            game.player.refresh(full = True)
            unused_stuff = game.player['special equipment']
            healing_charges = min(game.player['max health'] - game.player['health'], unused_stuff)
            if healing_charges > 0:
                game.player['health'] += healing_charges
                message('You use your leftover equipment to bandage your wounds. You recover {0} health.'.format(healing_charges), 'health')
                unused_stuff -= healing_charges 
            if unused_stuff > 0:
                stamina_charges = min(game.player['max health'] - game.player['health'], unused_stuff)
                if stamina_charges > 0:
                    game.player['stamina'] += stamina_charges
                    message('You trade in the rest of your leftover equipment to enjoy some refreshments. You recover {0} stamina.'.format(stamina_charges), 'stamina')
            if unused_stuff < 10:
                game.player['special equipment'] = 10                    
                message('After your heroic deed, the townspeople are eager to show their gratitude. You restock your special equipment.', 'wooden')
        
        self.ui.render_hud()
        if game.player['stamina'] == 0:
            message("You're too fatigued to resist falling into a nightmare-filled sleep. You have to face the Nightmare Tyrant right now!", 'warning')
            infotip("You're too fatigued to resist falling into a nightmare-filled sleep. You have to face the Nightmare Tyrant right now!", 'light nightmare')
                                        
        """ Briefing stage - choosing next quest """
        quest_choices = []
        quest_choices.append(('Save & Quit [S]', 'S'))
        custom_index = ord('a')
        for quest in game.quests:
            
            quest_choices.append(("{0} {2}{1}{3}".format( quest.name, 
                                    'completed' if quest.is_completed() else chr(custom_index),
                                    '(' if quest.is_completed() else '[',
                                    ')' if quest.is_completed() else ']',
                                    ), chr(custom_index), 'completed quest' if quest.is_completed() else quest.name))
            custom_index += 1
        
        while True:
            chosen_quest_index = self.ui.menu('Which task shall you undertake next?',
                                    quest_choices, indexing = self.ui.MENU_NUMCOMBO_INDEX)
            if chosen_quest_index is not None:
                chosen_quest_index -= 1 # 
                if chosen_quest_index < 0: #>= len(game.quests): # save & quit
                    game.save(self.savefile)
                    quit_game = True
                    break
                chosen_quest = game.quests[chosen_quest_index]
                if chosen_quest.is_finished():
                    self.ui.msgbox("This quest has been already completed!", 'completed quest')
                elif game.player['stamina'] == 0 and chosen_quest.name != 'Nightmare World':
                    self.ui.msgbox("You're too fatigued to resist falling into a nightmare-filled sleep. You have to face the Nightmare Tyrant right now!", 'light nightmare', margins = True)
                else: #quest properly chosen, can proceed to board game part
                    if chosen_quest.name == 'Nightmare World':
                        game.player = new_player()
                    break
                
        #generate map (at this point it's not drawn to the screen)
        if not quit_game:
            game.current_quest = chosen_quest 
            game.board = generate_board(width = 44, height = 44, player = game.player, mission = chosen_quest.name)
        return quit_game
        
    def board_play(self):
        quit_game = False
        finish_tactical_play = False
        self.initialize_fov()
        turns_taken = 0
        center_on_player = True
        self.refresh = True
        game.state = 'playing'
        infotip(str(game.board[game.player.x, game.player.y])[5:])
        while not quit_game and not finish_tactical_play:
            #render the screen
            if self.refresh:
                if game.current_quest.is_finished():
                    infotip("Your current quest is completed! You can leave by pressing [L] if there are no opponents in sight.", 'completed quest')
                self.ui.render_all(center_on_player)
                self.refresh = False
            else:
                self.ui.render_mouseover_info()
            libtcod.console_flush()
            if turns_taken == 0 or game.state != 'playing':
                turns_taken, quit_game, center_on_player = self.player_action(center_on_player)
            if game.current_quest is not None:
                game.current_quest.progress_objectives()
                if game.current_quest.is_finished():
                    if game.current_quest == game.quests[-1] and not game.state.startswith('victory'): #TODO: do something better than game.quests[-1]
                        message("Congratulations! You have defeated the Nightmare Tyrant and saved your town!", 'gold')
                        infotip("Press <Escape> to return to main menu.", 'gold')
                        game.state = 'victory'
                        self.ui.render_all(center_on_player)
                        self.ui.msgbox("Congratulations! You have saved your town! Press any key to return to main menu.", default_text_color = 'gold')
                        quit_game = True
                        
            if game.state == 'debriefing':
                finish_tactical_play = True
            if game.state == 'playing' and turns_taken > 0:
                game.fov_check()
                if game.agents_in_fov == [game.player]:
                    """ Tactical refresh - check if an encounter is over (no opponent in sight) and remove all tactical effects,
                        both from player character and their opponents. 
                    """                        
                    if game.player.refresh(full = False):
                        message("Without anything threatening you, you can catch your breath, readjust your outfit, " \
                                " reload your crossbow and you're ready for another fight!", 'positive')
                        
                infotip(str(game.board[game.player.x, game.player.y])[5:])                
                turns_taken -= 1
                for agent in game.agents:
                    if agent != game.player and agent['ai'] is not None:
                        agent.act()
                        if game.state == 'dead':
                            break
            if quit_game:
                if game.state in ['dead','victory']:
                    quests_completed = sum([1 for quest in game.quests if quest.is_completed()])
                    score = 1 if sum(game.kill_list.values()) > 0 else 0 
                    score += (10 - quests_completed) if game.state == 'victory' else quests_completed       
                    self.achievements.add_highscore(score)
                    self.ui.msgbox("Your score: {0}/10".format(score), default_text_color = 'gold')
                    self.achievements.remember(game)
                    #os.remove(self.savefile) TODO: enable to disable savescumming. Possibly make a debug option
                    game.state = 'finished'
                #else:
                #    game.save(self.savefile)
        return quit_game
    
    def play(self):
        self.prepare_play()
        quit_game = False
        while not libtcod.console_is_window_closed():
            if game.state.find('briefing') >= 0:
                quit_game = self.text_play()
            if not quit_game:
                quit_game = self.board_play() or game.state == 'victory'
            if quit_game:
                break
            
        if game.state != 'finished':
            game.save(self.savefile)

    def select_direction(self, unblocked_only = False):
        """return a tile next to the player selected by a direction key
            or None if escaped
        """
        infotip('Select a direction ([qweadzxc] or Numpad [12346789]) or press' \
                '<Esc>, <Space>, or Right Click to cancel.'.format())
        while True:
            self.ui.render_hud()
            libtcod.console_flush()
            player_action = input_manager.command(self.ui.get_input())
            if player_action is not None and player_action.startswith('DIRECTION '):
                dx, dy = directions[player_action[10:]]
                #the coordinates the player is moving to/attacking
                x = game.player.x + dx
                y = game.player.y + dy
                if not game.board.contains_position(x,y):
                    infotip("There's nothing relevant in that direction!", 'warning')
                else: #try to find an attackable object there
                    return game.board[x,y]
            elif player_action in ['CANCEL', 'QUIT']:
                return None

    def select_target_tile(self, occupied = False, max_range = None, in_fov = True, confirm_action = None):
        """return a tile left-clicked in player's FOV (optionally in a range),
           or None if right-clicked or escaped.
        """
        infotip('Left click on a{0} to select a target or press <{1}> to select closest applicable target.' \
                '<Esc>, <Space>, or Right Click to cancel.'.format('n enemy' if occupied else ' tile', input_manager.key_of(confirm_action)))
        while True:
            #TODO: refresh stuff under mouse and hud
            self.ui.render_hud()
            libtcod.console_flush()
            player_action = input_manager.command(self.ui.get_input())
            tile = self.ui.get_tile_under_mouse()
            if player_action == 'SELECT' and tile is not None:
                fov_constraint = not in_fov or game.in_fov(tile.x, tile.y)
                range_constraint = max_range is None or distance(game.player.x, game.player.y, tile.x, tile.y) <= max_range
                occupied_constraint = not occupied or tile.occupant is not None #TODO: and not player?  
                if fov_constraint and range_constraint and occupied_constraint:
                    if occupied:
                        return tile.occupant
                    else:
                        return tile
                elif not fov_constraint:
                    infotip("Target must be visible!")
                elif not range_constraint:
                    infotip("Target outside range of {0} tiles.".format(max_range))
                elif not occupied_constraint:
                    infotip("Target must be a combatant!")
            elif confirm_action is not None and player_action == confirm_action:
                """TODO: the goal is to confirm manual (or automatic default) selection of a tile/occupant targeted,
                    but for now we'll just select one automatically.
                """  
                default_target = closest_agent(game.player.x, game.player.y, max_range if max_range is not None else game.player['sight'])
                if default_target is not None and not occupied:
                    return game.board[default_target.x, default_target.y]
                return default_target
            elif player_action in ['CANCEL', 'QUIT']:
                return None
            
    def player_action(self, old_center_on_player):
        """ Returns how many turns passed, whether the player has quit
            and whether the camera should center on player.
            
            @return turns taken, quit game, center camera on player
            
            #TODO: consider a dictionary of player action outcomes.
            #TODO: consider returning the amount of turns an action would take (NEW: eh? don't we do it already?)
        """
        center_on_player = old_center_on_player
        player_action = input_manager.command(self.ui.get_input())
        if player_action is None:
            return (0, False, center_on_player)
        player = game.player        
        center_on_player = True
        self.refresh = True
        """
            TODO: REALTIME decisions for hasty players
            if not game.turn_mode: 
                if self.action_timeout == 0:
                    self.action_timeout = GAME_ACTION_TIMEOUT
                    return 'waiting'
                self.action_timeout -= 1
        """
        if game.state == 'playing':
            """ To avoid excessive returning group the actions:
                1st group uses up a turn
                2nd group after else doesn't.
            """
            turns_taken = 1 #default action length is one, but lets not get constrained by it!
            if player_action == 'QUIT':
                return (0, True, center_on_player)
            if player_action.startswith('DIRECTION '):
                self.player_act_in_direction(directions[player_action[10:]])
            elif player_action == 'INTERACT':
                """ This allows to interact with features that can't be bumped into.
                """
                traversible_features = []
                x,y = player.x, player.y
                for tile in game.board.ring(x,y):
                    if tile.feature is not None and not tile.is_blocked():
                        traversible_features.append((tile.feature,
                                                    directions['full names'][(tile.x -x, tile.y - y)]))
                if len(traversible_features) > 1:
                    self.ui.msgbox("Too many to choose!") #TODO: let them choose
                elif len(traversible_features) == 1:
                    traversible_features[0][0].interact(player)
                else:
                    infotip("Nothing to interact with.")
                    turns_taken = 0
            elif player_action.startswith('PICK UP'):
                #TODO: menu to choose a pickup if more than one
                tile = game.board[player.x, player.y] 
                if tile.items:
                    for item in [item for item in tile.items]:
                        player.pick_up(item)
                else:
                    infotip("Nothing to pick up.")
                    turns_taken = 0
            elif player_action.startswith('LEAVE'):
                if game.current_quest is not None and not game.current_quest.is_finished():
                    infotip("You have to finish your quest or die trying!",'failure')
                elif game.agents_in_fov!=[game.player]:
                    infotip("You can't leave during a fight!",'warning')
                else:
                    game.state = 'debriefing'
                turns_taken = 0
            elif player_action.startswith('DROP'):
                chosen_item = self.inventory_menu('Press the key next to an item to drop it, or any other to cancel.\n', ['droppable'])
            elif player_action.startswith('INVENTORY'):
                chosen_item = self.inventory_menu('Press the key next to an item to equip it, or any other to cancel.\n')
                if chosen_item is not None:
                    player.equip(chosen_item)
            elif player_action.startswith('USE'):
                action_type = player_action[4:]
                item = player[action_type + ' item']
                if item is not None and (not item.has_flag('limited') or player['special equipment'] > 0):
                    if item[action_type].requires_target:
                        if item.has_flag('close range'):
                            infotip("You prepare to deploy your {0}...".format(item.name))
                            target = self.select_direction(unblocked_only = item.has_flag('deployed'))
                        else:
                            infotip("You take aim with your {0}...".format(item.name))
                            target = self.select_target_tile(occupied = (action_type == 'ranged'), confirm_action = player_action)
                        if target is None:
                            infotip("Use of " + item.name + " cancelled.")
                            return (0, False, center_on_player)
                        item.use(game.player, action_type, target)
                    else:
                        item.use(game.player, action_type)
                else:
                    turns_taken = 0
                    if item.has_flag('limited') and player['special equipment'] == 0:
                        infotip("Not enough special equipment to use {0}!".format(articled(item.name)), 'special equipment')
                    else:
                        infotip("Nothing to use in " + player_action[4:] + " slot!")
            elif player_action.startswith('SELECT'):
                tile = self.ui.get_tile_under_mouse()
                if tile is not None:
                    if tile.occupant is not None and player['ranged item'] is not None and tile.visible:
                        player['ranged item'].use(game.player, 'ranged', tile.occupant)
                    else:
                        #TODO: Go (move) to tile
                        if player['ranged item'] is None:
                            message('You point your finger at {1} {0} and say: "Pif paf! You`re dead, {0}!"'.format( \
                                    'darkness' if not tile.visible else tile.occupant.name if tile.occupant is not None else tile.terrain.name, \
                                    a_or_an(tile.occupant.name) if tile.occupant is not None else 'the'
                                    ), 'joke')
                        else :
                            infotip("Shoot only at enemies, ricochets are dangerous!")
                        return (0, False, center_on_player)
                else:
                    #TODO: handle HUD elements selection
                    return (0, False, center_on_player)
            else :
                if player_action.startswith('CAMERA '):
                    dx, dy = directions[player_action[7:]]
                    self.ui.move_camera(dx * 10, dy * 10)
                    center_on_player = False
                elif player_action == 'DISPLAY MESSAGES':
                    self.ui.listbox('Message log:', game.msg_archive, width = INFO_WIDTH, scroll_to_bottom = True)
                elif player_action == 'DISPLAY KILL LIST':
                    kill_list = []
                    for kill, killcount in game.kill_list.items():
                        kill_list.append("{0} {1}{2}".format(killcount, kill, 's' if killcount > 1 else ''))
                    self.ui.listbox('Kill list:', kill_list, width = INFO_WIDTH)
                elif player_action == 'DISPLAY CHARACTER INFO':
                    self.ui.msgbox('TODO: Display character info.')
                elif player_action == 'HELP':
                    self.ui.msgbox('Keybindings: \n' \
                                   'Movement: \n' \
                                   '     [q] [w] [e] [a] [d] [z] [x] [c]\n' \
                                   '     numpad digits\n' \
                                   '     arrow keys (only orthogonally)\n' \
                                   'Wait: numpad 5 or [s]\n' \
                                   'Move Camera: ctrl + arrow keys \n' \
                                   'Save & Quit: ESC\n' \
                                   'Close doors: Space\n' \
                                   'Pick up items: [g]\n' \
                                   'Open inventory: [i]\n' \
                                   'Shoot: [r] or LMB on an enemy\n' \
                                   'Use equipped medkit: [v]\n' \
                                   'Throw equipped grenade: [f]\n' \
                                   'Activate equipped lightning launcher: [tab]\n' \
                                   'Show quest log: [Q]\n' \
                                   'Show this help: [?]'
                                   )
                elif player_action == 'QUEST LOG':
                    objective_list = []
                    for objective in game.current_quest.objectives.values():
                        objective_text = objective.description
                        objective_text += '.' if objective.repeats == 1 else ': {0}/{1}'.format(objective.accomplished, objective.repeats)
                        objective_color = 'objective '+ objective.status
                        if debug.info or objective.is_available():
                            objective_list.append((objective_text, objective_color))
                    self.ui.listbox(game.current_quest.name, objective_list)
                return (0, False, center_on_player)
            return (turns_taken, False, center_on_player)
        elif game.state == 'dead':
            if player_action in ['QUIT', 'INTERACT']:
                return (0, True, center_on_player)
        return (0, False, center_on_player) #TODO: does it make sense

    def inventory_menu(self, header, filters = []):
        #show a menu with each item of the inventory as an option
        filter_set = set(filters)
        filtered_inventory = [item for item in game.player.inventory if filter_set.issubset(set(item.flags))]
        if len(filtered_inventory) == 0:
            options = ['Nothing applicable.']
            #TODO: make it a msgboxy inventory
        else:
            options = []
            for item in filtered_inventory:
                option = item.name 
                for flag in ['ranged', 'melee', 'consumable', 'deployable', 'activated']:
                    if game.player[flag + ' item'] == item:
                        option += ' (equipped)' 
                options.append(option)
        index = self.ui.menu(header, options)
        #if an item was chosen, return it
        if index is None or len(filtered_inventory) == 0: return None
        return filtered_inventory[index]
    
    def player_act_in_direction(self, direction):
        """ Move or attack or perform other contextual actions.
            TODO: move to Agent class?
        """
        dx, dy = direction
        #the coordinates the player is moving to/attacking
        x = game.player.x + dx
        y = game.player.y + dy
        if not game.board.contains_position(x,y):
            infotip("It is the edge of the world!")
            return
        #try to find an attackable object there
        target_tile = game.board[x,y]
        if target_tile.is_blocked():
            if target_tile.occupant is not None:
                if target_tile.occupant.is_player():
                    message("You wait.")
                else:
                    game.player.melee(target_tile.occupant)
            elif target_tile.feature is not None:
                target_tile.feature.interact(game.player)
            else:
                infotip("You can't go there! {0} is in the way.".format(target_tile.blocking_piece.name.capitalize()))
        else:
            game.player.move_to(x, y)
            game.fov_recompute = True
    
    def prepare_play(self):
        """ This method:
            1) computes the necessary game data that cannot be handled by 
            game_data module to avoid circular dependencies 
            2) sets other things necessary for the game to be played that
            it would be a waste to do elsewhere.  
        """
        pass
        
    def initialize_fov(self):
        game.fov_recompute = True
        board = game.board
        game.fov_map = libtcod.map_new(board.width, board.height)
        for y in range(board.height):
            for x in range(board.width):
                game.set_fov_at(x, y)
        if not debug.fog:  #debug mode, reveal everything
            for y in range(board.height):
                for x in range(board.width):
                    board[x,y].explored = True
            for piece in game.pieces:
                piece.always_visible = True

        